# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["router"]

import asyncio
import contextlib
import contextvars
import http
import logging
import mimetypes
import os
import pathlib
import posixpath
import re
import urllib.parse
from typing import Annotated

import fastapi.responses
import packaging.version
import pooch  # pyright: ignore[reportMissingTypeStubs]
import pydantic
import pydantic_extra_types.semantic_version  # noqa: TC002
import requests

from . import _core

router = fastapi.APIRouter(route_class=_core.APIRoute)


@router.get("/python/pymanager/{name}")
async def get_python_install_manager_for_windows(
    name: Annotated[
        str,
        fastapi.Path(pattern=r"^python-manager-.+\.msix?$"),
    ],
) -> fastapi.Response:
    if not re.fullmatch(
        packaging.version.VERSION_PATTERN,
        name[15 : name.endswith(".msi") - 5],
        re.VERBOSE | re.IGNORECASE,
    ):
        return fastapi.Response(status_code=404)
    ctx = _core.context.get()
    urls = [
        urllib.parse.unquote(str(url)).format(version="pymanager", name=name)
        for url in ctx["config"].upstream.python
    ]
    return await _core.stream(urls)


@router.get("/python/{version}/{name}")
async def get_embedded_python(
    version: pydantic_extra_types.semantic_version.SemanticVersion,
    name: str,
) -> fastapi.Response:
    if (
        version < "3.5.0"
        or version.prerelease is not None
        or version.build is not None
        or name not in {
            f"python-{version}-embed-amd64.zip",
            f"python-{version}-embed-arm64.zip",
            f"python-{version}-embed-win32.zip",
        }
    ):
        return fastapi.Response(status_code=404)
    ctx = _core.context.get()
    urls = [
        urllib.parse.unquote(str(url)).format(version=version, name=name)
        for url in ctx["config"].upstream.python
    ]
    media_type, _ = mimetypes.guess_type(name)
    return await _core.stream(urls, media_type=media_type)


@router.get(
    "/python-build-standalone/{tag}/{name}",
    dependencies=_core.immutable,
)
async def get_standalone_python(
    tag: Annotated[
        str,
        fastapi.Path(min_length=8, max_length=8, pattern=r"^\d+$"),
    ],
    name: str,
) -> fastapi.Response:
    media_type, _ = mimetypes.guess_type(name)
    cache_location = pathlib.Path("python-build-standalone", tag, name)
    async with contextlib.AsyncExitStack() as stack:
        if await _core.cached_or_locked(cache_location, stack):
            return fastapi.responses.FileResponse(
                cache_location,
                media_type=media_type,
            )
        ctx = _core.context.get()
        escaped = name.replace("+", "%2B")  # Required by aliyun mirrors
        urls = [
            posixpath.join(str(url), tag, escaped)
            for url in ctx["config"].upstream.python_build_standalone
        ]
        sha256, size = await _get_standalone_python_sha256_and_size(tag, name)
        return await _core.stream(
            urls,
            media_type=media_type,
            stack=stack,
            cache_location=cache_location,
            sha256=sha256,
            size=size,
        )
    return _core.unreachable()


@router.get("/uv/python-downloads.json")
async def get_uv_python_downloads_json() -> fastapi.Response:
    ctx = contextvars.copy_context()
    lock = ctx[_core.context]["locks"]["python-downloads.json"]
    ctx.run(_core.cache_action.set, "cache-or-fetch")
    async with contextlib.AsyncExitStack() as stack:
        await stack.enter_async_context(lock)
        return await asyncio.create_task(
            _core.stream(
                "https://api.github.com/repos/astral-sh/uv/contents/crates/uv-python/download-metadata.json",
                headers={
                    "Accept": "application/vnd.github.raw+json",
                    "X-GitHub-Api-Version": "2022-11-28",
                },
                stack=stack,
            ),
            context=ctx,
        )


async def _get_standalone_python_sha256_and_size(
    tag: str,
    name: str,
) -> tuple[bytes, int | None]:
    try:
        release = await _core.GitHubRelease.fetch(
            f"python-build-standalone/{tag}.json",
            owner="astral-sh",
            repo="python-build-standalone",
            tag_name=tag,
        )
    except requests.RequestException:
        pass
    except OSError, pydantic.ValidationError:
        _logger.exception("Failed to get GitHub release metadata")
    else:
        for asset in release.assets:
            if asset.name == name:
                if sha256 := asset.sha256():
                    return sha256, asset.size
                break
        else:
            raise fastapi.HTTPException(404)
    cache_location = pathlib.Path("python-build-standalone", tag, "SHA256SUMS")
    loop = asyncio.get_running_loop()
    async with _core.cached_or_locked(cache_location) as cached:
        if not cached:
            ctx = _core.context.get()
            dir_, fname = os.path.split(cache_location)
            urls = [
                posixpath.join(str(url), tag, "SHA256SUMS")
                for url in ctx["config"].upstream.python_build_standalone
            ]
            for url in _core.load_balance(urls):
                with contextlib.suppress(Exception):
                    await loop.run_in_executor(
                        None,
                        pooch.retrieve,  # pyright: ignore[reportUnknownArgumentType, reportUnknownMemberType]
                        url,
                        None,
                        fname,
                        dir_,
                    )
                    break
            else:
                status_code = http.HTTPStatus.GATEWAY_TIMEOUT
                raise fastapi.HTTPException(status_code)
    return await loop.run_in_executor(
        None,
        _parse_standalone_python_sha256,
        cache_location,
        f"  {name}\n".encode("ascii"),
    )


def _parse_standalone_python_sha256(
    cache_location: pathlib.Path,
    name: bytes,
) -> tuple[bytes, None]:
    with cache_location.open("rb") as f:
        for line in f:
            if line.endswith(name):
                return bytes.fromhex(line[:64]), None  # pyright: ignore[reportArgumentType]
    raise fastapi.HTTPException(404)


_logger = logging.getLogger("mahoraga")



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["router"]

import asyncio
import contextlib
import contextvars
import logging
import mimetypes
import pathlib
import posixpath

import fastapi
import httpx
import pydantic

from mahoraga import _core

router = fastapi.APIRouter(route_class=_core.APIRoute)


@router.get("/{name}")
async def get_uv_github_release(name: str) -> fastapi.Response:
    ctx = _core.context.get()
    media_type, _ = mimetypes.guess_type(name)
    tag, sha256, size = await _get_distribution_metadata(name)
    if tag and sha256:
        cache_location = pathlib.Path("uv", tag, name)
        headers = {"Cache-Control": "public, max-age=31536000, immutable"}
        async with contextlib.AsyncExitStack() as stack:
            if await _core.cached_or_locked(cache_location, stack):
                return fastapi.responses.FileResponse(
                    cache_location,
                    headers=headers,
                    media_type=media_type,
                )
            urls = [
                posixpath.join(str(url), tag, name)
                for url in ctx["config"].upstream.uv.tag
            ]
            response = await _core.stream(
                urls,
                headers=headers,
                media_type=media_type,
                stack=stack,
                cache_location=cache_location,
                sha256=sha256,
                size=size,
            )
            if not httpx.codes.is_error(response.status_code):
                return response
    urls = [
        posixpath.join(str(url), name)
        for url in ctx["config"].upstream.uv.latest
    ]
    return await _core.stream(urls, media_type=media_type)


async def _get_distribution_metadata(
    name: str,
) -> tuple[str | None, bytes | None, int | None]:
    ctx = contextvars.copy_context()
    ctx.run(_core.cache_action.set, "cache-or-fetch")
    try:
        async with ctx[_core.context]["locks"]["uv/latest.json"]:
            raw = await asyncio.create_task(
                _core.get(
                    "https://api.github.com/repos/astral-sh/uv/releases/latest",
                    headers=_core.headers,
                ),
                context=ctx,
            )
    except fastapi.HTTPException:
        pass
    else:
        try:
            rel = _core.GitHubRelease.model_validate_json(raw)
        except pydantic.ValidationError:
            _logger.exception("Failed to get GitHub release metadata")
        else:
            for asset in rel.assets:
                if asset.name == name:
                    return rel.tag_name, asset.sha256(), asset.size
            raise fastapi.HTTPException(404)
    if name == "dist-manifest.json":
        return None, None, None
    return await _get_distribution_metadata_from_manifest(name)


async def _get_distribution_metadata_from_manifest(
    name: str,
) -> tuple[str | None, bytes | None, int | None]:
    ctx = _core.context.get()
    urls = [
        posixpath.join(str(url), "dist-manifest.json")
        for url in ctx["config"].upstream.uv.latest
    ]
    try:
        raw = await _core.get(urls)
    except fastapi.HTTPException:
        return None, None, None
    try:
        manifest = _DistributionManifest.model_validate_json(raw)
    except pydantic.ValidationError:
        _logger.exception("Failed to get uv distribution manifest")
        return None, None, None
    try:
        artifact = manifest.artifacts[name]
    except KeyError as e:
        raise fastapi.HTTPException(404) from e
    tag = manifest.announcement_tag
    if hexdigest := artifact.checksums.sha256:
        return tag, bytes.fromhex(hexdigest), None
    sha256 = f"{name}.sha256"
    if sha256 not in manifest.artifacts:
        return tag, None, None
    urls = [
        posixpath.join(str(url), tag, sha256)
        for url in ctx["config"].upstream.uv.tag
    ]
    try:
        raw = await _core.get(urls)
    except fastapi.HTTPException:
        return tag, None, None
    return tag, bytes.fromhex(raw[:64]), None  # pyright: ignore[reportArgumentType]


class _Checksums(pydantic.BaseModel, extra="ignore"):
    sha256: str | None = None


class _Artifact(pydantic.BaseModel, extra="ignore"):
    checksums: _Checksums = _Checksums()


class _DistributionManifest(pydantic.BaseModel, extra="ignore"):
    announcement_tag: str
    artifacts: dict[str, _Artifact]


_logger = logging.getLogger("mahoraga")



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["parselmouth", "router", "split_repo"]

import fastapi

from mahoraga import _core

from . import _packages, _repodata, _sharded_repodata
from ._parselmouth import router as parselmouth
from ._sharded_repodata import split_repo

router = fastapi.APIRouter(route_class=_core.APIRoute)
router.include_router(_repodata.router)
router.include_router(_sharded_repodata.router)
router.include_router(_packages.router)  # Must be the last included



# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# /// script
# dependencies = ["py-rattler >=0.20.0"]
# requires-python = ">=3.14"
# ///

"""Install Pixi.

For details, see https://hingebase.github.io/mahoraga/tutorial.html#pixi
"""

import argparse
import asyncio
import os
import shutil
from pathlib import Path
from typing import NoReturn

import rattler.networking
import rattler.rattler


async def _install_pixi(
    target_prefix: os.PathLike[str],
    cache_dir: os.PathLike[str] | None,
    mahoraga_base_url: str,
    version: rattler.VersionSpec,
) -> None:
    specs = [rattler.MatchSpec(f"pixi {version}", strict=True)]
    client = rattler.Client([
        rattler.networking.MirrorMiddleware({
            "https://conda.anaconda.org/": [
                mahoraga_base_url.rstrip("/") + "/conda/",
            ],
        }),
    ])
    records = await rattler.solve(
        channels=["conda-forge"],
        specs=specs,
        gateway=rattler.Gateway(cache_dir, client=client),
        virtual_packages=rattler.VirtualPackage.detect(),
    )
    await _rattler_install(
        records,
        target_prefix,
        cache_dir,
        show_progress=False,
        client=client,
        requested_specs=specs,
    )


def _main() -> NoReturn:
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "mahoraga_base_url",
        nargs="?",
        default="http://127.0.0.1:3450/",
        help="\b",
    )
    parser.add_argument(
        "-v", "--version",
        default="*",
        type=rattler.VersionSpec,
        help="Pixi version",
    )
    args = parser.parse_args()

    # https://pixi.sh/latest/reference/environment_variables/#configurable-environment-variables
    pixi_home = _pixi_home()
    cache_dir = _pixi_cache_dir()

    asyncio.run(
        _install_pixi(
            pixi_home,
            cache_dir,
            args.mahoraga_base_url,
            args.version,
        ),
    )
    pixi_home = pixi_home.resolve(strict=True)
    if pixi := shutil.which("pixi", path=pixi_home / "bin"):
        os.environ["PIXI_HOME"] = str(pixi_home)
        if cache_dir:
            os.environ["PIXI_CACHE_DIR"] = str(cache_dir.resolve(strict=True))
        os.execv(pixi, ("pixi", "info"))  # noqa: S606

    # Most likely a permisson error if reaching here
    # For conciseness, error handling is omitted
    raise OSError


def _pixi_cache_dir() -> Path | None:
    env = os.environ
    if cache_dir := env.get("PIXI_CACHE_DIR") or env.get("RATTLER_CACHE_DIR"):
        return Path(cache_dir)
    if cache_dir := env.get("XDG_CACHE_HOME"):
        try:
            last_resort = Path(cache_dir, "pixi").resolve(strict=True)
            if last_resort.is_dir():
                return last_resort
        except OSError:
            pass
    return None


def _pixi_home() -> Path:
    if pixi_home := os.getenv("PIXI_HOME"):
        return Path(pixi_home)
    return Path.home() / ".pixi"


async def _rattler_install(  # noqa: PLR0913, PLR0917
    records: list[rattler.RepoDataRecord],
    target_prefix: str | os.PathLike[str],
    cache_dir: os.PathLike[str] | None = None,
    installed_packages: list[rattler.PrefixRecord] | None = None,
    reinstall_packages: set[str] | None = None,
    ignored_packages: set[str] | None = None,
    platform: rattler.Platform | None = None,
    execute_link_scripts: bool = False,  # noqa: FBT001, FBT002
    show_progress: bool = True,  # noqa: FBT001, FBT002
    client: rattler.Client | None = None,
    requested_specs: list[rattler.MatchSpec] | None = None,
) -> None:
    await rattler.rattler.py_install(  # pyright: ignore[reportUnknownMemberType]
        records=records,
        target_prefix=str(target_prefix),
        cache_dir=cache_dir,
        installed_packages=installed_packages,
        reinstall_packages=reinstall_packages,
        ignored_packages=ignored_packages,
        platform=platform and platform._inner,  # noqa: SLF001  # pyright: ignore[reportPrivateUsage, reportUnknownMemberType]
        client=client and client._client,  # noqa: SLF001  # pyright: ignore[reportPrivateUsage, reportUnknownMemberType]
        execute_link_scripts=execute_link_scripts,
        show_progress=show_progress,
        requested_specs=requested_specs and [
            spec._match_spec  # noqa: SLF001  # pyright: ignore[reportPrivateUsage, reportUnknownMemberType]
            for spec in requested_specs
        ],
    )


if __name__ == "__main__":
    _main()
